import javax.media.opengl.GL;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.glu.GLU;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

public class HoughRenderer implements GLEventListener, KeyListener {
   private double rotateT = 0.0f;
   private static final GLU glu = new GLU();
 
   private double[][] lines;
   private int[][][] image;
   private int[][][][] parameterImage;
   private int tempValue;
   private int tempMaxValue;
   private final int tempSleep = 30;
   private int tempSleepCounter = 0;
   private int parameterImageRange;
   private double[] vv1 = new double[3];
   private double[] vv2 = new double[3];
   private double[] vv3 = new double[3];
   private double[] vv4 = new double[3];
   private double[] vv5 = new double[3];
   private double[] vv6 = new double[3];
   private double[] vv7 = new double[3];
   private double[] vv8 = new double[3];
   private boolean printMode = true;

   public void setLines(double[][] lines) {
	   this.lines = lines;
	   
	   // convert to degrees
	   for (double[] line : this.lines) {
		   line[0] = 360*line[0]/(2*Math.PI);
		   line[1] = 360*line[1]/(2*Math.PI);
		   line[3] = 360*line[3]/(2*Math.PI);
	   }
   }
   
   public void setImage(int[][][] image) {
	   this.image = image;
   }
   
   public void setParameterImage(int[][][][] parameterImage, int range) {
	   this.tempMaxValue = parameterImage[0][0][0].length;
	   this.parameterImage = parameterImage;
	   this.parameterImageRange = range;
   }
   
   public void display(GLAutoDrawable gLDrawable) {
       final GL gl = gLDrawable.getGL();
       int windowWidth = gLDrawable.getWidth();
       int windowHeight = gLDrawable.getHeight();
       
       gl.glClear(GL.GL_COLOR_BUFFER_BIT);
       gl.glClear(GL.GL_DEPTH_BUFFER_BIT);
       
       // Image space viewport
       gl.glViewport(0, 0, windowWidth / 2, windowHeight);
       gl.glLoadIdentity();
       gl.glRotated(-90.0, 1.0, 0.0, 0.0);
       gl.glTranslated(0.0, 100.0, -10.0);
       gl.glRotated(rotateT, 0.0, 0.0, 1.0); // z-axis rotation animation
       drawAxes(gl, 100.0);
       drawImage(gl);
       drawLines(gl, 500.0);
       
       // Parameter space viewport
       gl.glViewport(windowWidth / 2, 0, windowWidth / 2, windowHeight);
       gl.glLoadIdentity();
       gl.glRotated(-90.0, 1.0, 0.0, 0.0);
       gl.glTranslated(0.0, 100.0, -10.0);
       gl.glRotated(rotateT, 0.0, 0.0, 1.0); // z-axis rotation animation
       if (this.tempSleepCounter < this.tempSleep) {
    	   this.tempSleepCounter++;
       } else {
    	   this.tempSleepCounter = 0;
    	   this.tempValue = (this.tempValue + 1) % this.tempMaxValue;
       }
       drawAxes(gl, 100.0);
       drawParameterImage(gl); 
       
       rotateT += 0.02;
   }

   private void drawAxes(GL gl, double length) {
	   gl.glPointSize(3.0f);
	   if (this.printMode) {
		   gl.glColor3d(0.3, 0.3, 0.3);
	   } else {
		   gl.glColor3d(0.0, 0.0, 1.0);
	   }
	   gl.glLineWidth(0.2f);
	   
	   gl.glBegin(GL.GL_LINES);
       /*gl.glColor3d(1.0, 0.0, 0.0);*/ gl.glVertex3d(0.0, 0.0, 0.0);
       /*gl.glColor3d(1.0, 0.0, 0.0);*/ gl.glVertex3d(length, 0.0, 0.0);
       gl.glEnd();
       gl.glBegin(GL.GL_POINTS);
       for (int x=0; x<length; x++) {
    	   gl.glVertex3d((double)x, (double)0, (double)0);
       }
       gl.glEnd();
       
       gl.glBegin(GL.GL_LINES);
       /*gl.glColor3d(0.0, 1.0, 0.0);*/ gl.glVertex3d(0.0, 0.0, 0.0);
       /*gl.glColor3d(0.0, 1.0, 0.0);*/ gl.glVertex3d(0.0, length, 0.0);
       gl.glEnd();
       gl.glBegin(GL.GL_POINTS);
       for (int y=0; y<length; y++) {
    	   gl.glVertex3d((double)0, (double)y, (double)0);
       }
       gl.glEnd();
       
       gl.glBegin(GL.GL_LINES);
       /*gl.glColor3d(0.0, 0.0, 1.0);*/ gl.glVertex3d(0.0, 0.0, 0.0);
       /*gl.glColor3d(0.0, 0.0, 1.0);*/ gl.glVertex3d(0.0, 0.0, length);
       gl.glEnd();
       gl.glBegin(GL.GL_POINTS);
       for (int z=0; z<length; z++) {
    	   gl.glVertex3d((double)0, (double)0, (double)z);
       }
       gl.glEnd();
	   
   }
   
   private void drawImage(GL gl) {
	   if (this.image == null) {
		   return;
	   }
	   gl.glPushMatrix();
	   double centerX = (int)(this.image[0][0].length/2.0);
	   double centerY = (int)(this.image[0].length/2.0);
	   double centerZ = (int)(this.image.length/2.0);
	   gl.glTranslated(-centerX, -centerY, -centerZ);
	   gl.glPointSize(2.0f);
	   gl.glBegin(GL.GL_POINTS);
	   gl.glColor3d(1.0, 1.0, 1.0);
	   for (int z=0; z<this.image.length; z++) {
		   for (int y=0; y<this.image[z].length; y++) {
			   for (int x=0; x<this.image[z][y].length; x++) {
				   if (this.image[z][y][x] > 0) {
					   gl.glVertex3d((double)x, (double)y, (double)z);
					   this.drawVoxel(gl, x, y, z, 0.3, 0.8, 0.3);
				   }
			   }   
		   }   
	   }
	   gl.glEnd();
	   gl.glPopMatrix();
   }
   
   private void drawParameterImage(GL gl) {
	   if (this.parameterImage == null) {
		   return;
	   }
	   for (int psi=0; psi<parameterImage.length; psi++) {
		   for (int phi=0; phi<parameterImage[psi].length; phi++) {
			   for (int rho=0; rho<parameterImage[psi][phi].length; rho++) {
				   //for (int omega=0; omega<image2[psi][phi][this.image2temporalValue].length; omega++) {
					   int value = parameterImage[psi][phi][rho][this.tempValue];
					   if (value > 0) {
						   double intensity = 1-(double)value/(double)this.parameterImageRange;
						   this.drawVoxel(gl, (double)psi, (double)phi, (double)rho, intensity, intensity, intensity);
						   
					   }
				   //}
			   }   
		   }   
	   }
	   gl.glEnd();
   }
   
   private void drawVoxel(GL gl, double x, double y, double z, double red, double blue, double green) {
	   //gl.glPointSize(3.0f); 
	   
	   double voxelSize = 0.35;
	   gl.glColor3d(red, blue, green);
	   
	   vv1[0] = x-voxelSize; vv1[1] = y-voxelSize; vv1[2] = z-voxelSize;
	   vv2[0] = x+voxelSize; vv2[1] = y-voxelSize; vv2[2] = z-voxelSize;
	   vv3[0] = x+voxelSize; vv3[1] = y-voxelSize; vv3[2] = z+voxelSize;
	   vv4[0] = x-voxelSize; vv4[1] = y-voxelSize; vv4[2] = z+voxelSize;
	   
	   vv5[0] = x+voxelSize; vv5[1] = y+voxelSize; vv5[2] = z-voxelSize;
	   vv6[0] = x-voxelSize; vv6[1] = y+voxelSize; vv6[2] = z-voxelSize;
	   vv7[0] = x-voxelSize; vv7[1] = y+voxelSize; vv7[2] = z+voxelSize;
	   vv8[0] = x+voxelSize; vv8[1] = y+voxelSize; vv8[2] = z+voxelSize;
	   
	   gl.glBegin(GL.GL_QUAD_STRIP);
	   gl.glVertex3d(vv1[0], vv1[1], vv1[2]);
	   gl.glVertex3d(vv4[0], vv4[1], vv4[2]);
	   gl.glVertex3d(vv2[0], vv2[1], vv2[2]);
	   gl.glVertex3d(vv3[0], vv3[1], vv3[2]);
	   gl.glVertex3d(vv5[0], vv5[1], vv5[2]);
	   gl.glVertex3d(vv8[0], vv8[1], vv8[2]);
	   gl.glVertex3d(vv6[0], vv6[1], vv6[2]);
	   gl.glVertex3d(vv7[0], vv7[1], vv7[2]);
	   gl.glVertex3d(vv1[0], vv1[1], vv1[2]);
	   gl.glVertex3d(vv4[0], vv4[1], vv4[2]);
	   gl.glEnd();
	   
	   gl.glBegin(GL.GL_QUADS);
	   gl.glVertex3d(vv4[0], vv4[1], vv4[2]);
	   gl.glVertex3d(vv3[0], vv3[1], vv3[2]);
	   gl.glVertex3d(vv8[0], vv8[1], vv8[2]);
	   gl.glVertex3d(vv7[0], vv7[1], vv7[2]);
	   
	   gl.glVertex3d(vv6[0], vv6[1], vv6[2]);
	   gl.glVertex3d(vv5[0], vv5[1], vv5[2]);
	   gl.glVertex3d(vv2[0], vv2[1], vv2[2]);
	   gl.glVertex3d(vv1[0], vv1[1], vv1[2]);
	   gl.glEnd();
   }
   
   private void drawLines(GL gl, double length) {
	   if (this.lines == null) {
		   return;
	   }
	   
	   for (double[] line : this.lines) {
	       double psi = line[0];
	       double phi = line[1];
	       double rho = line[2];
	       double omega = line[3];
		       
	       gl.glPushMatrix();
	       
	       // rotate into place
	       gl.glRotated(phi, -Math.sin(2*Math.PI*psi/360), Math.cos(2*Math.PI*psi/360), 0.0);
	       gl.glRotated(psi, 0.0, 0.0, 1.0);
	       gl.glTranslated(0.0, 0.0, rho);
	       gl.glRotated(omega, 0.0, 0.0, Math.signum(rho)*1.0);
	              
	       // draw line
	       gl.glLineWidth(0.2f);
	       gl.glBegin(GL.GL_LINES);
	       gl.glColor3d(0.0, 0.0, 1.0);
	       gl.glVertex3d(0.0, -length/2.0, 0.0);
	       gl.glVertex3d(0.0, length/2.0, 0.0);
	       gl.glEnd();
	       
	       gl.glPopMatrix();

       }
   }
   
   public void displayChanged(GLAutoDrawable gLDrawable, boolean modeChanged, boolean deviceChanged) {
   }

   public void init(GLAutoDrawable gLDrawable) {
       final GL gl = gLDrawable.getGL();
       gl.glShadeModel(GL.GL_SMOOTH);
       if (this.printMode) {
    	   gl.glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
       } else {
    	   gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
       }
       gl.glClearDepth(1.0f);
       gl.glEnable(GL.GL_DEPTH_TEST);
       gl.glDepthFunc(GL.GL_LEQUAL);
       gl.glHint(GL.GL_PERSPECTIVE_CORRECTION_HINT, GL.GL_NICEST);
       gLDrawable.addKeyListener(this);
    }

   public void reshape(GLAutoDrawable gLDrawable, int x, int y, int width, int height) {
       final GL gl = gLDrawable.getGL();
       if(height <= 0) {
           height = 1;
       }
       final float h = (float)width / (float)height;
       gl.glMatrixMode(GL.GL_PROJECTION);
       gl.glLoadIdentity();
       glu.gluPerspective(50.0f, h, 1.0, 500.0);
       gl.glMatrixMode(GL.GL_MODELVIEW);
       gl.glLoadIdentity();
   }

   public void keyPressed(KeyEvent e) {
       if(e.getKeyCode() == KeyEvent.VK_ESCAPE) {
           HoughDisplayControl.bQuit = true;
           HoughDisplayControl.displayT.stop();
           System.exit(0);
       }
   }

   public void keyReleased(KeyEvent e) {
   }

   public void keyTyped(KeyEvent e) {
   }
}