%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Prototype file, demonstrating pruning on voxel grid skeleton
% Knut Vidar Siem, Spring 2007
%
%
  clear all;
% Parameters:
%
  inputFileName = 'input.img';
  inputDims = [256 517 256]; % Length of input, dimension 1
  outputFileName = 'output.img';
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Read data

disp '--- Pruning prototype ---';
data = rawRead(inputFileName, inputDims(1),inputDims(2),inputDims(3));
disp 'Read data.';

data = logical(data);
disp 'Data is now binary.';


%% Create structuring elements

s = zeros(3, 3, 3);
s(:,:,1) = [-1 -1 -1
            -1 -1 -1
            -1 -1 -1];
s(:,:,2) = [-1 -1 -1
            -1  1 -1
            -1 -1 -1];
s(:,:,3) = [-1  0  0
            -1  0 -1
            -1 -1 -1];


% Coordinate system:
%   _ (1)
%  |\
%    \o---> (3)
%     |
%     v (2)
%
% baseStrels indices illustrating 'dice' rolling according to dimensions
% 5
% 1 2 3 4
% 6

% rotate s three times about dimension 2
numBaseStrels = 6; % 6 faces
baseStrels = cell(numBaseStrels, 1);

baseStrels{1} = s;
baseStrels{2} = rot903d(baseStrels{1}, 2);
baseStrels{3} = rot903d(baseStrels{2}, 2);
baseStrels{4} = rot903d(baseStrels{3}, 2);

% rotate s 1 time about dimension 3 each way
baseStrels{5} = rot903d(s, 3);
baseStrels{6} = rot903d(rot903d(baseStrels{5}, 3), 3);  % rotate two more to "go the other way"


% rotate each of the base structuring elements about dimension 1
numUnmirroredStrels = numBaseStrels*4;
unmirroredStrels = cell(numUnmirroredStrels, 1);
for baseStrelIndex=1:numBaseStrels
    baseStrel = baseStrels{baseStrelIndex};
    unmirroredStrels{(baseStrelIndex-1)*4 + 1} = baseStrel;
    for rotations = 2:4
        currentIndex = (baseStrelIndex-1)*4 + rotations;
        unmirroredStrels{currentIndex} = rot903d(unmirroredStrels{currentIndex - 1}, 1);
    end
end
clear currentIndex;
clear rotations;
clear baseStrel;
clear baseStrelIndex;

clear baseStrels;
clear numBaseStrels;
%strels = unmirroredStrels;
%clear unmirroredStrels;
%clear numUnmirroredStrels;

clear s;


%% Test structuring elements

%testInput = zeros(9, 9, 9);
%middle1 = ceil(size(testInput,1)/2);
%middle2 = ceil(size(testInput,2)/2);
%middle3 = ceil(size(testInput,3)/2);

% 3-D cross
%testInput(:, middle2, middle3) = 1;
%testInput(middle1, :, middle3) = 1;
%testInput(middle1, middle2, :) = 1;

% Tilted 3-D cross (assume square)
%for i=1:size(testInput, 1)
%    testInput(i,i,i) = 1;
%    testInput(size(testInput,1)+1-i, i, i) = 1;
%    testInput(i, size(testInput,2)+1-i, i) = 1;
%    testInput(i, i, size(testInput,3)+1-i) = 1;
%end
%
%numStrels = size(strels, 1);
%work = testInput;
%numIterations = 1;
%for i=1:numIterations
%    for strelIndex = 1:numStrels
%       work = work & ~bwhitmiss(work, strels{strelIndex});
%    end
%end


%% Unnecessary because of symmetry
% mirror all structuring elements about any dimension (e.g. 1)
numStrels = numUnmirroredStrels*2;
strels = cell(numStrels, 1);
for unmirroredStrelIndex = 1:numUnmirroredStrels
    unmirroredStrel = unmirroredStrels{unmirroredStrelIndex};
    strels{(unmirroredStrelIndex-1)*2 + 1} = unmirroredStrel;
    strels{(unmirroredStrelIndex-1)*2 + 2} = flipdim(unmirroredStrel, 1);
end

% consistency check
%numIdenticalStrels = 0;
%for testStrelIndex = 1:numStrels
%   for strelIndex = 1:numStrels
%       if testStrelIndex == strelIndex
%           continue
%       else
%           if strels{testStrelIndex} == strels{strelIndex}
%               numIdenticalStrels = numIdenticalStrels +1;
%               disp 'Found identical structuring elements at:';
%               testStrelIndex
%               strelIndex
%           end
%       end
%   end
%end
%disp 'Number of identical structuring elements: ';
%numIdenticalStrels

%% Prune data set

numStrels = size(strels, 1);
work = data;
numIterations = 1;
for i=1:numIterations
    for strelIndex = 1:numStrels
       work = work & ~bwhitmiss(work, strels{strelIndex});
       disp 'Finished structuring element: ';
       strelIndex
    end
    disp '--- finished iteration ---';
    i
end

%% Write output file
output = work;
if islogical(output)
    output = output*127;
end
fid = fopen(outputFileName,'wb');
fwrite(fid, output);
fclose(fid);
disp 'File output complete.';

clear output;
